/*
	Ogg Vorbis plugin for SoundPlay.
	Copyright © 2000 Marco Nelissen (marcone@xs4all.nl)
	libvorbis copyright Monty <monty@xiph.org> and The XIPHOPHORUS Company
	http://www.xiph.org/

	This is an example of a data-producing plugin for SoundPlay.
	Permission is hereby granted to use this code for the development
	of other SoundPlay-plugins.
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <AppKit.h>
#include <InterfaceKit.h>
#include <MediaDefs.h>
#include <MediaRoster.h>
#include <TimeSource.h>
#include <Path.h>

#include "OggVorbisPlugin.h"
#include "VGroup.h"
#include "MSlider.h"

#define CONNECT(x)

/* ===========================================================
	Plugin-descriptor structure.
	This structure is used to gather information about a plugin
	and	instantiate and destroy specific instances of a plugin.
*/
static plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	"sp-vorbis",1, PLUGIN_IS_DECODER|PLUGIN_NEEDS_READER,

	"Ogg Vorbis Decoder",
	"By Marco Nelissen.\n\n"
	"Uses libvorbis by Xiphorus\n"
	"(See http://www.xiph.org)",
	NULL, //about,
	NULL, //configure,
	&getplugin,
	&destroyplugin
};

/* ===========================================================
	Plugin-operations structure.
	This is used to actually perform operations on a plugin.
	It is returned by the getplugin() function from the 
	plugin_descriptor structure above.
*/
static decoder_plugin_ops stream_ops={
	PLUGIN_DECODER_MAGIC,PLUGIN_DECODER_VERSION,

	NULL,	// open
	NULL,	// close
	info,
	read,
	NULL,	// play
	NULL,	// stop
	NULL,	// setvolume
	NULL,	// setpitchandspeed
	NULL, 	// seek,
	NULL,	// position,
	bufferamount
};

static decoder_plugin_ops file_ops={
	PLUGIN_DECODER_MAGIC,PLUGIN_DECODER_VERSION,

	NULL,	// open
	NULL,	// close
	info,
	read,
	NULL,	// play
	NULL,	// stop
	NULL,	// setvolume
	NULL,	// setpitchandspeed
	seek,
	position,
	NULL	// bufferamount
};


/* ===========================================================
	Return the list of plugins in this file.
*/
static plugin_descriptor *plugs[]={
	&plugin,
	0
};

plugin_descriptor **get_plugin_list(void)
{
	return plugs;
}

/* ===========================================================
	Return the list of filetypes supported by all the plugins
	in this file
*/
static supported_type supportedtypes[]=
{
	{"audio/x-vorbis",	"Ogg Vorbis file",	"Ogg Vorbis file",			"ogg",	NULL},
	{NULL,				NULL,				NULL,						NULL,	NULL}
};

supported_type *get_supported_types(void)
{
	return supportedtypes;
}


/* ===========================================================
	Implementation of the plugin_descriptor functions
*/

static void *getplugin(void **data, const char *name, const char *header, uint32 size, plugin_info *info, SoundPlayBufferedReader *reader)
{
	if(info->controller->Version() < 0x00049000)
		return NULL;

	VorbisReader *vr=NULL;

	vr = new VorbisReader(reader, info->controller);
	
	if(vr->InitCheck()==B_OK)
	{
		*data=vr;
		if(reader->IsRandomlyAccessible())
			return &file_ops;
		return &stream_ops;
	}
	delete vr;
	return NULL;
}

static void destroyplugin(void *ops,void *_ogg)
{
	VorbisReader *ogg=(VorbisReader*)_ogg;
	delete ogg;
}


/* ===========================================================
	Implementation of the decoder_plugin_ops functions
*/
static int32 read(void *_ogg, char *buf, ulong count)
{
	VorbisReader *ogg=(VorbisReader*)_ogg;
	return ogg->Read(buf,count);
}

static status_t info(void *_ogg, file_info *info)
{
	VorbisReader *ogg=(VorbisReader*)_ogg;

	vorbis_info *vorbisInfo = ov_info(&ogg->vorbishandle, -1);

// Fill in the file_info structure
	info->bitrate=ov_bitrate(&ogg->vorbishandle,-1);
	if(info->bitrate<=0)
		info->bitrate = vorbisInfo->bitrate_nominal;
	info->samplerate=vorbisInfo->rate;
	info->numchannels=vorbisInfo->channels;
	strcpy(info->name,"");
	sprintf(info->typedesc,"Ogg Vorbis %g kHz, %d kbps",float(vorbisInfo->rate)/1000, int(info->bitrate/1000));
	strcpy(info->mimetype,"audio/x-vorbis");
	info->granularity=4096;
	if(ogg->reader->IsRandomlyAccessible())
		info->framecount=long(ov_time_total(&ogg->vorbishandle, -1)*info->samplerate);
	else
		info->framecount=long(0x7fffffff);

	info->samplesize=2;
	if(B_HOST_IS_LENDIAN)
		info->byteorder=B_LITTLE_ENDIAN;
	else
		info->byteorder=B_BIG_ENDIAN;
	info->sampleformat=B_LINEAR_SAMPLES;
	if(!ogg->reader->IsRandomlyAccessible())
		info->flags=PLUGIN_FILELENGTH_UNKNOWN;
	
	return B_OK;
}

static status_t seek(void *_ogg, uint32 pos)
{
	VorbisReader *ogg=(VorbisReader*)_ogg;
	
	int ret = ov_pcm_seek(&ogg->vorbishandle, pos);
//	printf("seek(%d): %d\n",pos, ret);
	return ret;
}

static uint32 position(void *_ogg)
{
	VorbisReader *ogg=(VorbisReader*)_ogg;
	int32 pos = ov_pcm_tell(&ogg->vorbishandle);
//	printf("now at: %d\n",pos);
	return pos;
}

float bufferamount(void *_ogg)
{
	VorbisReader *ogg=(VorbisReader*)_ogg;
	return ogg->BufferAmount();
}

// =====================================
// "file" I/O callback functions 
// =====================================

size_t vb_read(void *buffer, size_t numchunks, size_t chunksize, void *io)
{
//	printf("callback reading data\n");
	SoundPlayReader *reader=(SoundPlayReader*)io;
	return reader->Read(buffer,numchunks*chunksize);
}

int vb_seek(void *io, ogg_int64_t where, int whence)
{
//	printf("seek callback called (%Ld, %d)\n", where, whence);
	SoundPlayReader *reader=(SoundPlayReader*)io;
	if(reader->IsRandomlyAccessible())
		return reader->Seek(where, whence);
	return -1;
}

int vb_close(void *io)
{
	return 0; // we'll dispose of the SoundPlayReader elsewhere;	
}

long vb_ftell(void *io)
{
	SoundPlayReader *reader=(SoundPlayReader*)io;
//	printf("ftell callback called\n");
	return reader->Position();
}

// =====================================
// vorbisreader class 
// =====================================

VorbisReader::VorbisReader(SoundPlayBufferedReader *reader, SoundPlayController *ctrl)
{
	initcheck=B_ERROR;
	this->reader=reader;
	controller = ctrl;
   	memset(&vorbishandle,0,sizeof(vorbishandle));

	ov_callbacks callbacks =
	{
		(size_t (*)(void *, size_t, size_t, void *))  vb_read,
		(int (*)(void *, ogg_int64_t, int))           vb_seek,
		(int (*)(void *))                             vb_close,
		(long (*)(void *))                            vb_ftell
	};
	if(ov_open_callbacks((void *)reader, &vorbishandle, NULL, 0, callbacks)<0)
		return;
	initcheck=B_OK;
	currentsection=0;
}

float VorbisReader::BufferAmount()
{
	return reader->BufferPercentage();
}

VorbisReader::VorbisReader(const char *name)
{
	initcheck=B_ERROR;
	reader=NULL;
   	memset(&vorbishandle,0,sizeof(vorbishandle));
	FILE *file=fopen(name, "r");
	if(!file)
		return;
	if(ov_open(file, &vorbishandle, NULL, 0) < 0)
		return;
	initcheck=B_OK;
	currentsection=0;
}

status_t VorbisReader::InitCheck()
{
	return initcheck;
}

VorbisReader::~VorbisReader()
{
	ov_clear(&vorbishandle);
}


int32 VorbisReader::Read(char *buf, ulong numframes)
{
// Ogg Vorbis returns variable length blocks of data, and very short ones at that.
// In order to do reliable backwards-playback we need to return at least as much
// data as SoundPlay asks for.
// We will occasionally return more data than asked for, but this doesn't matter
// (there is extra space after the buffer pointed to by 'buf')
	uint32 total=0;
	int32 ret;
	uint32 zerocounter=0;
	vorbis_info *vorbisInfo = ov_info(&vorbishandle, -1);
	int32 framesize = 2*vorbisInfo->channels;
	do {
		ret=ov_read(&vorbishandle,buf,numframes*framesize,0,2,1,&currentsection);
		
		BString title;
		BString artist;
		char **comments = vorbishandle.vc->user_comments;
		for(int i=0;i<vorbishandle.vc->comments;i++)
		{
			if(strncmp("title=",comments[i],6)==0)
				title = comments[i]+6;
			if(strncmp("artist=",comments[i],7)==0)
				artist = comments[i]+7;
		}
		
		BString metadata;

		if(artist != "")
		{
			metadata << "StreamTitle='" << artist;
			if(title != "")
			{
				metadata << " - " << title << "';";
			}
		}
		else if(title != "")
		{
			metadata << "StreamTitle='" << title << "';";
		}
		if(metadata != "" && metadata != lastmeta)
		{
			lastmeta = metadata;
			controller->SetMetaData(lastmeta.String(),this);
		}
		
		if(ret > 0)
		{
			buf+=ret;
			total+=ret/framesize;
		}
		else
			zerocounter++;
	} while((ret||zerocounter<10) && total <= numframes);
	return total;
}



